//
// FILE: vline.cpp
//

// OVERVIEW:
//
// This file takes as input a v-line file and a cnf file containing a mapping
// the v-line file's variables' numbering and the original variables' 
// numbering and produces a v-line file of true variables for the original
// cnf file.  The cnf file is generated by the preprocessor and has a special
// comment block at the beginning which contains the information necessary to
// generate a v-line for the original cnf.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

namespace hypre {

#define _DEBUG

enum state {TRUE, FALSE, EQV, NEWNUM};

void usage() {
  printf("Usage: vline [input vline file] [input cnf file]\n"); 
  exit(1);
}

#ifdef WIN32
int __cdecl main(int argc, char *argv[]) {
#else 
int main(int argc, char *argv[]) {
#endif

  int i, retval, lit, orig_var, new_var;
  bool value;
  char * vline_filename;
  char * cnf_filename;
  int OLDNVARS, NEWNVARS;
  FILE * vfp;
  FILE * cfp;
  FILE * outfp;
  char ch;
  char word[81];
  bool * assignments;
  int * varNumber;
  state * varState;
  int * newNum;

  if(argc == 1)
    usage();

  vline_filename = argv[1];
  if((vfp = fopen(vline_filename, "r")) == NULL) {
    printf("ERROR: Could not open file named %s.\n", vline_filename);
    usage();
  }

  cnf_filename = argv[2];
  if((cfp = fopen(cnf_filename, "r")) == NULL) {
    printf("ERROR: Could not open file named %s.\n", cnf_filename);
    usage();
  }

  // Find the line in the cnf file that contains a the number of variables
  // in the original theory and the number of variables in the reduced 
  // theory.  It begins with a "c *"
  while((retval = fscanf(cfp, "c %c", &ch)) && (ch != '*')) {
    if((retval == EOF) || feof(cfp)) {
      printf("ERROR: Could not find line with variable numbers in cnf file.\n");
      fclose(vfp);
      fclose(cfp);
      exit(0);
    }
    while(fscanf(cfp, "%c", &ch) && (ch != '\n'))
      ;
  }
  
  // After the * are two numbers.  The first is the number of variable in
  // the original and the second is the number in the reduced theory.
  fscanf(cfp, "%d %d", &OLDNVARS, &NEWNVARS);

  // Skip the rest of the white space on this line.
  while(fscanf(cfp, "%c", &ch) && (ch != '\n'));

  // Set up data structures to store information about the number mapping.
  varNumber = new int[OLDNVARS+1];
  varState = new state[OLDNVARS+1];
  newNum = new int[OLDNVARS+1];
  assignments = new bool[NEWNVARS+1];


  // Now process the v-line file containing the variable truth assignment
  // for the reduced theory.

  // Find the line containing the variable assignments.  It starts with the
  // letter 'v'.
  while((retval = fscanf(vfp, "%c", &ch)) && (ch != 'v')) {
    if((retval == EOF) || feof(vfp)) {
      printf("ERROR: Could not find the v-line in the file.\n");
      fclose(vfp);
      fclose(cfp);
      exit(0);
    }
    while(fscanf(vfp, "%c", &ch) && (ch != '\n'))
      ;
  }

  for(i = 1; i <= NEWNVARS; i++) {
    fscanf(vfp, "%d", &lit);
    
#ifdef _DEBUG
    if(abs(lit) != i)
      printf("ERROR: Literal missing in v-line file!\n");
#endif // _DEBUG

    if(lit > 0)
      assignments[i] = true;
    else
      assignments[i] = false;
  }

#ifdef _DEBUG
  fscanf(vfp, "%d", &lit);
  if(lit != 0)
    printf("ERROR: More variables than expected in v-line file!\n");
#endif // _DEBUG

  // Create output v-line file which contains the matching truth assignments 
  // using the variable numbering from the original cnf file.
  
  if((outfp = fopen("vline.out", "w")) == NULL) {
    printf("ERROR: Could not open file vline.out.\n");
    exit(1);
  }

  // Now, go back to looking through the cnf file and find and process
  // the 

  // Find mapping of original cnf numbering to new cnf numbering inside
  // the new cnf file.
  while((retval = fscanf(cfp, "c %c", &ch)) && (ch != '*')) {
    if((retval == EOF) || feof(cfp)) {
      printf("ERROR: Could not find mapping in the cnf file.\n");
      fclose(vfp);
      fclose(cfp);
      exit(0);
    }
    while(fscanf(cfp, "%c", &ch) && (ch != '\n'))
      ;
  }

  // Now begin to read the mapping from the old cnf file and store the
  // mapping information in two arrays.
  i = 1;
  while((retval = fscanf(cfp, " c %c", &ch)) && (ch != '*')) {
    if((retval == EOF) || feof(vfp)) {
      printf("ERROR: cnf file ended prematurely.\n");
      fclose(vfp);
      fclose(cfp);
      exit(0);
    }
    
    // Skip lines with negative literals.
    while(fscanf(cfp, "%c", &ch) && (ch != '\n'));

    fscanf(cfp, "c %d %s", &orig_var, word);
    varNumber[i] = orig_var;

    //temp
    //fprintf(outfp, "%d ", orig_var);

    if(strcmp(word, "T") == 0)
      varState[i] = TRUE;
    else if(strcmp(word, "F") == 0)
      varState[i] = FALSE;
    else if(strcmp(word, "equivalent") == 0) {
      fscanf(cfp, " to %d", &new_var);
      varState[i] = EQV;
      newNum[i] = new_var;
    }
    // Arbitrarily assign variables whose truth value did not affect the 
    // satisfiability of the theory to TRUE.
    else if(strcmp(word, "not") == 0) {
      varState[i] = TRUE;
      // Skip the rest of this line.
      while(fscanf(cfp, "%c", &ch) && (ch != '\n'));
    }
    else {
      if((new_var = strtol(word, NULL, 10)) == 0) {
	printf("ERROR: Expected integer but did not find one.\n");
	exit(1);
      }
      varState[i] = NEWNUM;
      newNum[i] = new_var;
    }
    i++;
  }

  //temp
  //fprintf(outfp, "\n");

  // Now print out the v-line file for the old cnf's numbering of the
  // variables.
  fprintf(outfp, "v ");

  for(i = 1; i <= OLDNVARS; i++) {
    if(varState[i] == TRUE)
      fprintf(outfp, "%d ", varNumber[i]);
    else if(varState[i] == FALSE)
      fprintf(outfp, "-%d ", varNumber[i]);
    else if(varState[i] == EQV) {
      if(varState[abs(newNum[i])] == NEWNUM) {
	if(newNum[i] > 0)
	  value = assignments[newNum[newNum[i]]];
	else
	  value = !assignments[newNum[abs(newNum[i])]];
	
	if(value)
	  fprintf(outfp, "%d ", varNumber[i]);
	else
	  fprintf(outfp, "-%d ", varNumber[i]);
      }
      else if(((newNum[i] > 0) && (varState[newNum[i]] == TRUE))
	      || ((newNum[i] < 0) && (varState[abs(newNum[i])] == FALSE)))
	fprintf(outfp, "%d ", varNumber[i]);
      else if(((newNum[i] < 0) && (varState[abs(newNum[i])] == TRUE))
	      || ((newNum[i] > 0) && (varState[newNum[i]] == FALSE)))
	fprintf(outfp, "-%d ", varNumber[i]);
      else {
	printf("ERROR: Ran into unexpected situation.\n");
	exit(1);
      }
    }
    else if(varState[i] == NEWNUM) {
      if(assignments[newNum[i]])
	fprintf(outfp, "%d ", varNumber[i]);
      else
	fprintf(outfp, "-%d ", varNumber[i]);
    }
    
#ifdef _DEBUG
    else
      printf("ERROR: varState[%d] has an unknown value!\n", i);
#endif // _DEBUG

  }

  fprintf(outfp, "0");

  fclose(vfp);
  fclose(cfp);
  fclose(outfp);
}

}
